@@ -1,5 +1,5 @@
 from abc import ABC, abstractmethod
-from inspect import signature
+from inspect import signature, isclass
 from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                     List, Optional, overload, Sequence, Set, Tuple, Type,
                     TYPE_CHECKING, TypeVar, Union)
@@ -11,21 +11,18 @@ from cirq.ops.gate_features import SingleQubitGate, TwoQubitGate
 
 from quple import QuantumCircuit, TemplateCircuitBlock
 from quple.components.interaction_graphs import interaction_graph
-from quple.utils.utils import merge_pqc
+from quple.utils.utils import merge_pqc, get_unique_symbols
 
 
 class ParameterisedCircuit(QuantumCircuit, ABC):
     """Parameterised Quantum Circuit (PQC)
     
-    For construction of parameterised quantum circuit with strongly entangling
-    circuit architecture referenced from the circuit-centric classifier design
-    https://arxiv.org/pdf/1804.00633.pdf. 
-    
-    In this architecture, the parameterised circuit is composed of L copies of
-    a primary circuit block. Each circuit block consists of a layer of single
-    qubit gates (the rotation layer) applied to each of the n qubits, followed
-    by a layer of two (or three) qubit gates (the entanglement layer) to entangle
-    all input qubits with a given interaction graph.
+    The `ParameterisedCircuit` architecture consists of alternating rotation and entanglement 
+    layers that are repeated for a certain number of times. In both layers, parameterized 
+    circuit-blocks act on the circuit in a defined way. The rotation layer consists of single 
+    qubit gate operations (rotation blocks) that are applied to every qubit in the circuit. 
+    The entanglement layer consists of two (or multiple) qubit gate operations (entanglement blocks) 
+    applied to the set of qubits defined by an interaction graph.
     
     Common single qubit gates in the rotation layer include the Hadamard gate,
     Pauli X, Pauli Y, Pauli Z gate and the corresponding rotation gates
@@ -86,6 +83,11 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                 parameter_symbol:str='θ',
                 final_rotation_layer:bool=False,
                 flatten_circuit:bool=False,
+                reuse_param_per_depth:bool=False,
+                reuse_param_per_layer:bool=False,
+                reuse_param_per_template:bool=False,
+                parameter_index:Optional[int]=None,
+                parameter_scale=1,
                 name:str='ParameterisedCircuit',
                 *args, **kwargs):
         """ Creates a parameterised circuit following the scheme of 
@@ -98,8 +100,9 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                    Number of times the layers are repeated.
                 rotation_blocks: A list of single qubit gate operations to be applied in the rotation layer.
                 entanglement_blocks: A list of multi qubit gate operations to be applied in the entanglement layer.
-                entangle_strategy: default='full'
+                entangle_strategy: default=None
                     Determines how the qubits are connected in an entanglement block.
+                    If None, it defaults to using full entanglement.
                     If str, it specifies the name of the strategy.
                     If callable, it specifies the function to map to an interaction graph.
                     If list of str, it specifies the names of a list of strategies. The 
@@ -114,6 +117,14 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                    Whether to add an extra final rotation layer to the circuit.
                 flatten_circuit: boolean, default=False
                    Whether to flatten circuit parameters when the circuit is modified.
+                reuse_param_per_depth: boolean, default=False
+                   Whether to reuse parameter symbols at every new depth (symbol starting index reset to 0)
+                reuse_param_per_layer: boolean, default=False
+                   Whether to reuse parameter symbols at every new layer (symbol starting index reset to 0)
+                reuse_param_per_template: boolean, default=False
+                   Whether to reuse parameter symbols at every new template block (symbol starting index reset to 0)
+                parameter_index: int, default=None
+                   Starting index of the first parameter
                 name: str
                    Name of the circuit. 
         """
@@ -122,7 +133,12 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         self._parameters = np.array([], dtype=object)
         self._readout_qubit = None
         self._flatten_circuit = flatten_circuit
-        self._entangle_strategy = entangle_strategy
+        self._entangle_strategy = entangle_strategy if entangle_strategy else 'full'
+        self._parameter_index = parameter_index
+        self._reuse_param_per_depth    = reuse_param_per_depth
+        self._reuse_param_per_layer    = reuse_param_per_layer
+        self._reuse_param_per_template = reuse_param_per_template
+        self._parameter_scale = parameter_scale        
         self.build(rotation_blocks, entanglement_blocks, entangle_strategy, copies,
                   final_rotation_layer)
     
@@ -135,15 +151,25 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         """Returns the readout qubit of the circuit
         """
         return self._readout_qubit
-            
+    
+    @property
+    def parameter_index(self):
+        """Returns the starting index for the next new parameter symbol
+        """
+        return self._parameter_index
+    
     @property
     def num_param(self):
         """Number of parameters in the circuit created through the 
            add rotation layer and add entanglement layer methods
         """
         return len(self._parameters)
-
-    def new_param(self, size:int=1) -> sp.Symbol:
+    
+    @property
+    def parameter_scale(self):
+        return self._parameter_scale
+    
+    def new_param(self, size:int=1, as_array=False) -> sp.Symbol:
         """Returns a new parameter(s) with updated array index(indices)
         
         Args:
@@ -153,13 +179,27 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         Returns:
             New parameter(s) with updated array index(indices)
         """
-        start = self.num_param
+        if size < 1:
+            raise ValueError('Number of new parameters created must be greater than zero')
+            
+        if self.parameter_index is None:
+            start = self.num_param
+        else:
+            start = self.parameter_index
+            self._parameter_index += size
+            
         params = np.array([sp.Symbol('%s_%s' % (self.parameter_symbol, i)) \
                           for i in range(start, start+size)])
-        self._parameters = np.append(self._parameters, params)
-        if len(params) == 1:
-            return params[0]
-        return params
+        all_params = np.append(self._parameters, params)
+        self._parameters = np.array(get_unique_symbols(all_params))
+        
+        if (len(params) == 1) and (not as_array):
+            return params[0]*self.parameter_scale
+
+        return params*self.parameter_scale
+    
+    def reset_index(self, index:int=0):
+        self._parameter_index = index        
         
     @property
     def parameters(self):
@@ -209,6 +249,12 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                     not isinstance(block, gate_feature):
                     raise ValueError('Gate operation {} should be a subclass of '
                                      '{}'.format(type(block), gate_feature))
+            elif isclass(block) and issubclass(block, cirq.Gate):
+                parameters = ParameterisedCircuit._get_gate_parameters(block)
+                n_param = len(parameters)
+                dummpy = sp.symarray('x',n_param)
+                kwargs = {parameters[i]:dummpy[i] for i in range(n_param)}
+                ParameterisedCircuit._validate_blocks([block(**kwargs)], gate_feature)
             elif callable(block):
                 n_param = ParameterisedCircuit._get_parameter_count(block)
                 if not n_param:
@@ -216,7 +262,16 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                                      'circuit block: {}'.format(block))
                 dummy_params = tuple(sp.symarray('x',n_param))
                 ParameterisedCircuit._validate_blocks([block(*dummy_params)], gate_feature)
-        
+    @staticmethod
+    def _get_gate_parameters(gate):
+        parameters = []
+        parameters_dict = dict(signature(gate).parameters)
+        for param_name in parameters_dict:
+            annotation = parameters_dict[param_name].annotation
+            if (annotation == sp.Basic) or (hasattr(annotation, '__args__') and sp.Basic in annotation.__args__):
+                parameters.append(param_name)
+        return parameters
+      
     @staticmethod
     def _get_parameter_count(gate_expr) -> int:
         '''Returns the number of parameters required for a gate operation expression
@@ -224,9 +279,9 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         to a cirq.Gate instance, a function that maps to cirq.Gate or a TemplateCircuitBlock instance
         '''
         if isinstance(gate_expr, cirq.Gate):
-            if isinstance(gate_expr, (cirq.XXPowGate, cirq.YYPowGate, cirq.ZZPowGate)):
-                return 1
             return 0
+        elif isclass(gate_expr) and issubclass(gate_expr, cirq.Gate):
+            return len(ParameterisedCircuit._get_gate_parameters(gate_expr))
         elif isinstance(gate_expr, TemplateCircuitBlock):
             return gate_expr.num_params
         elif callable(gate_expr):
@@ -316,6 +371,8 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
                 return interaction_graph[strategy[block_index]](self.n_qubit, num_block_qubits)
             elif all(callable(strat) for strat in strategy):
                 return strategy[block_index](self.n_qubit, num_block_qubits)
+            elif all(isinstance(strat, tuple) for strat in strategy):
+                return strategy
         else:
             raise ValueError('invalid entangle strategy: {}'.format(strategy))
         
@@ -329,16 +386,22 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
             gate: The gate operation to be parameterised
         Returns:
             A parameterised gate operation
-        """
+        """        
         if isinstance(gate, cirq.Gate):
             if isinstance(gate, (cirq.XXPowGate, cirq.YYPowGate, cirq.ZZPowGate)):
                 param = self.new_param()
                 return gate**param
             return gate
+        elif isclass(gate) and issubclass(gate, cirq.Gate):
+            parameters = ParameterisedCircuit._get_gate_parameters(gate)
+            n_param = len(parameters)
+            params = self.new_param(size=n_param, as_array=True)
+            kwargs = {parameters[i]:params[i] for i in range(n_param)}
+            return gate(**kwargs)    
         else:
             n_param = ParameterisedCircuit._get_parameter_count(gate)
             params = self.new_param(size=n_param)
-            params = tuple(params) if isinstance(params, np.ndarray) else (params,)
+            params = tuple(params) if isinstance(params, np.ndarray) else (params,)          
             return gate(*params)
         
     def add_rotation_layer(self, rotation_blocks:Optional[Union[str, cirq.Gate, Callable,
@@ -351,9 +414,13 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         """
         rotation_blocks = self._parse_rotation_blocks(rotation_blocks)
         for i, block in enumerate(rotation_blocks):
+            if self._reuse_param_per_layer:
+                self.reset_index()
             if isinstance(block, TemplateCircuitBlock):
                 interaction_graphs = self.get_interaction_graphs(i, block.num_block_qubits)
                 for qubits in interaction_graphs:
+                    if self._reuse_param_per_template:
+                        self.reset_index()
                     block.build(self, qubits)
             else:
                 for qubit in range(self.n_qubit):
@@ -382,10 +449,14 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         entangle_strategy = entangle_strategy or 'full' 
         entanglement_blocks = self._parse_entanglement_blocks(entanglement_blocks)
         for i, block in enumerate(entanglement_blocks):
+            if self._reuse_param_per_layer:
+                self.reset_index()            
             if isinstance(block, TemplateCircuitBlock):
                 interaction_graphs = self.get_interaction_graphs(i, block.num_block_qubits,
                                                                 entangle_strategy)
                 for qubits in interaction_graphs:
+                    if self._reuse_param_per_template:
+                        self.reset_index()                    
                     block.build(self, qubits)
             else:
                 interaction_graphs = self.get_interaction_graphs(i, 2, entangle_strategy)
@@ -438,11 +509,14 @@ class ParameterisedCircuit(QuantumCircuit, ABC):
         """
         self.clear()
         for _ in range(copies):
+            if self._reuse_param_per_depth:
+                self.reset_index()
             self.add_rotation_layer(rotation_blocks)
             self.add_entanglement_layer(entanglement_blocks, entangle_strategy)
             
         if final_rotation_layer:
             self.add_rotation_layer(rotation_blocks)
+            
         if self.flatten_circuit:
             self.flatten()                    
     
