@@ -141,8 +141,8 @@ class QuantumCircuit(cirq.Circuit):
             If None, defaults to quantum simulator.
     """
     def __init__(self, n_qubit:Union[int, Sequence[GridQubit]]=0, name:str='QuantumCircuit',
-                insert_strategy:InsertStrategy=None,
-                backend=None) -> None:
+                 insert_strategy:InsertStrategy=None,
+                 backend=None) -> None:
         """Creates a quantum circuit
         
         Args:
@@ -228,6 +228,34 @@ class QuantumCircuit(cirq.Circuit):
         if not isinstance(insert_strategy, InsertStrategy):
             raise ValueError('insert strategy must be derived from the cirq.circuits.InsertStrategy class')
         self._insert_strategy = insert_strategy
+        
+    def get_gate_operation(self, operation:[str, cirq.Gate], qubit_expr, params:Optional[Dict]=None):
+        """Obtain a list of gate operations on designated qubits
+        
+        Args:
+            operation: the gate operation to apply
+            qubit_expr: the qubits on which the gate operation is applied
+            params: a dict containing the parameters for the gate operation
+        
+        Returns:
+            List of gate operations on designated qubits
+        """
+        operation = QuantumCircuit._parse_gate_operation(operation)
+        if params is not None:
+            operation = operation(**params)
+        qubit_sequence = self.qr.get(qubit_expr)
+        if not isinstance(qubit_sequence, (list, np.ndarray)):
+            qubit_sequence = [qubit_sequence]
+        
+        gate_ops = None
+        if all(isinstance(qubits, GridQubit) for qubits in qubit_sequence):
+            gate_ops = [operation(qubit) for qubit in qubit_sequence]
+        elif all(isinstance(qubits, tuple) for qubits in qubit_sequence):
+            gate_ops = [operation(*qubits) for qubits in qubit_sequence]
+        else:
+            raise ValueError('Inconsistent qubit representation: {}'.format(qubit_sequence))
+
+        return gate_ops
 
     def apply_gate_operation(self, operation:[str, cirq.Gate], qubit_expr, params:Optional[Dict]=None):
         """Apply a gate operation on designated qubits
@@ -237,19 +265,9 @@ class QuantumCircuit(cirq.Circuit):
             qubit_expr: the qubits on which the gate operation is applied
             params: a dict containing the parameters for the gate operation
         """
-        operation = QuantumCircuit._parse_gate_operation(operation)
-        if params is not None:
-            operation = operation(**params)
-        qubit_sequence = self.qr.get(qubit_expr)
-        if not isinstance(qubit_sequence, (list, np.ndarray)):
-            qubit_sequence = [qubit_sequence]
+        gate_ops = self.get_gate_operation(operation, qubit_expr, params)
         strategy = self.insert_strategy
-        if all(isinstance(qubits, GridQubit) for qubits in qubit_sequence):
-            self.append([operation(qubit) for qubit in qubit_sequence], strategy=strategy)
-        elif all(isinstance(qubits, tuple) for qubits in qubit_sequence):
-            self.append([operation(*qubits) for qubits in qubit_sequence], strategy=strategy)
-        else:
-            raise ValueError('Inconsistent qubit representation: {}'.format(qubit_sequence))
+        self.append(gate_ops, strategy=strategy)
             
     def get_param_resolver(self, param_values: Dict):
         """
