@@ -6,7 +6,7 @@ from quple import TemplateCircuitBlock
 class RISWAPBlock(TemplateCircuitBlock):
     
     @staticmethod
-    def RYY(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int], theta):
+    def RYY(circuit:'quple.ParameterisedCircuit', theta, qubits:Sequence[int]):
         circuit.RX(np.pi/2, list(qubits))
         circuit.CX(tuple(qubits))
         circuit.RZ(theta, qubits[1])
@@ -14,7 +14,7 @@ class RISWAPBlock(TemplateCircuitBlock):
         circuit.RX(-np.pi/2, list(qubits))
         
     @staticmethod
-    def RXX(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int], theta):
+    def RXX(circuit:'quple.ParameterisedCircuit', theta, qubits:Sequence[int]):
         circuit.H(list(qubits))
         circuit.CX(tuple(qubits))
         circuit.RZ(theta, qubits[1])
@@ -24,9 +24,9 @@ class RISWAPBlock(TemplateCircuitBlock):
 
     def build(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int]):
         theta = circuit.new_param()
-        RISWAPBlock.RXX(theta, qubits)
-        RISWAPBlock.RYY(theta, qubits)
+        RISWAPBlock.RXX(circuit, theta, qubits)
+        RISWAPBlock.RYY(circuit, theta, qubits)
     
     @property
     def num_block_qubits(self) -> int:
-        return 2
+        return 2
\ No newline at end of file
