@@ -1,4 +1,4 @@
-from typing import Optional, Sequence, List, Dict, Union
+from typing import Optional, Sequence, List, Dict, Union, Callable
 from abc import ABC
 import logging
 
@@ -65,9 +65,9 @@ class VQC(tf.keras.Sequential):
                  repetitions=None,
                  loss='mse', 
                  activation='sigmoid',
-                 metrics=['binary_accuracy'],
+                 metrics=['binary_accuracy', 'qAUC'],
                  readout=None,
-                 trainable_dense_layer:bool=False,
+                 classical_layer:bool=False,
                  random_state:int=None,                 
                  name:str=None, *arg, **args):
         """ Creates a variational quantum classifier
@@ -96,8 +96,8 @@ class VQC(tf.keras.Sequential):
                 List of metrics to be evaluated by the model during training and testing. 
             readout: `cirq.PauliSum` or Python `list` of `cirq.PauliSum` objects
                 Measurement operators (observables) for the variational circuit layer.
-            trainable_dense_layer: boolean; default=False
-                Whether to train the weights in the the classical output layer.
+            classical_layer: boolean; default=False
+                Whether to add a classical output layer.
             random_state: Optional int
                 The random state for reproducible result.
             name: Optional str
@@ -121,19 +121,48 @@ class VQC(tf.keras.Sequential):
         self.repetitions = repetitions
         self.activation = activation        
         self.random_state = random_state
-        self.trainable_dense_layer = trainable_dense_layer
-        layers = self._get_vqc_layers()
+        self.classical_layer = classical_layer
+        layers = self._get_sequential_layers()   
         super(VQC, self).__init__(layers, name)
         self.compile(loss=loss, optimizer=optimizer, metrics=metrics)
         
-    def compile(self, optimizer='adam', loss='binary_crossentropy', metrics=['binary_accuracy'], *args, **kwargs):
-        super(VQC, self).compile(optimizer, loss, metrics, *args, **kwargs)
-        self._attributes['optimizer'] = self.optimizer._name if hasattr(self.optimizer, '_name') \
-                                        else self.optimizer.__name__
-        self._attributes['metrics'] = [metric._name if hasattr(metric, '_name')\
-                                       else metric.__name__ for metric in self.metrics]
-        self._attributes['loss'] = [loss.name if hasattr(loss, 'name') \
-                                    else loss.__name__ for loss in self.loss_functions]
+    @tf.function
+    def custom_auc(self, y_true, y_pred):
+        y_pred_ = (y_pred+1)/2
+        return self.auc_metric(y_true, y_pred_)        
+        
+    def compile(self, optimizer='adam', loss='binary_crossentropy', metrics=['binary_accuracy', 'qAUC'], *args, **kwargs):
+        _metrics = []
+        for metric in metrics:
+            if metric.lower() == 'qauc':
+                from quple.interface.tfq.metrics.metrics import qAUC
+                _metrics.append(qAUC())
+            else:
+                _metrics.append(metric)
+        super(VQC, self).compile(optimizer, loss, _metrics, *args, **kwargs)
+        # attribute checks for compatibility with older version of tensorflow
+        if hasattr(self.optimizer, '_name'):
+            optimizer_name = self.optimizer._name
+        elif hasattr(self.optimizer, '__name__'):
+            optimizer_name = self.optimizer.__name__
+        else:
+            raise ValueError('unable to extract optimizer name, please check tensorflow version')
+        if hasattr(self, 'compiled_metrics'):
+            metrics_name = [metric if isinstance(metric, str) else metric.name for metric in self.compiled_metrics._metrics]
+        elif hasattr(self, 'metrics'):
+            metrics_name = [metric.__name__ for metric in self.metrics]
+        else:
+            raise ValueError('unable to extract metrics name, please check tensorflow version')
+        if hasattr(self, 'compiled_loss'):
+            loss_name = self.compiled_loss._losses
+        elif hasattr(self, 'loss_functions'):
+            loss_name = [loss.__name__ for loss in self.loss_functions]
+        else:
+            raise ValueError('unable to extract loss name, please check tensorflow version')  
+
+        self._attributes['optimizer'] = optimizer_name
+        self._attributes['metrics']   = metrics_name
+        self._attributes['loss']      = loss_name
         
     @property
     def attributes(self):
@@ -153,7 +182,14 @@ class VQC(tf.keras.Sequential):
         self._encoding_circuit = val
         self._attributes['feature_dimension'] = feature_dimension
         self._attributes['encoding_circuit'] = val.name if isinstance(val, quple.QuantumCircuit) else ''
-        self._attributes['encoding_map'] = val.encoding_map.__name__ if isinstance(val, EncodingCircuit) else ''
+        encoding_map_name = ''
+        if isinstance(val, EncodingCircuit):
+            encoding_map = val.encoding_map
+            if hasattr(encoding_map, '__name__'):
+                encoding_map_name = val.encoding_map.__name__
+            elif hasattr(type(encoding_map), '__name__'):
+                encoding_map_name = type(val.encoding_map).__name__
+        self._attributes['encoding_map'] = encoding_map_name
         logger.info('Registered encoding circuit with feature dimension: {}'.format(feature_dimension))
     
     @property
@@ -234,7 +270,7 @@ class VQC(tf.keras.Sequential):
         self._attributes['activation'] = val.__name__ if val else ''
         self._activation = val
         
-    def _get_vqc_layers(self):
+    def _get_sequential_layers(self):
         
         input_layer = tf.keras.layers.Input(shape=(), dtype=tf.string)
         pqc_layer = tfq.layers.PQC(self.variational_circuit,
@@ -242,18 +278,18 @@ class VQC(tf.keras.Sequential):
                                    repetitions=self.repetitions,
                                    differentiator=self.differentiator,
                                    regularizer=self.regularizer)
-        if self.trainable_dense_layer:
+        layers = [input_layer, pqc_layer]
+        if self.classical_layer:
+            #output_layer = tf.keras.layers.Dense(1, activation=self.activation, 
+            #                kernel_initializer=tf.keras.initializers.Constant(value=1)) 
+            #output_layer.trainable = False
             output_layer = tf.keras.layers.Dense(1, activation=self.activation)
-        else:
-            output_layer = tf.keras.layers.Dense(1, activation=self.activation, 
-                            kernel_initializer=tf.keras.initializers.Constant(value=1)) 
-            output_layer.trainable = False
-    
-        return [input_layer, pqc_layer, output_layer]
+            layers.append(output_layer)
+        return layers
     
     def _reset_layers(self):
         self._layers = None
-        for layer in self._get_vqc_layers():
+        for layer in self._get_sequential_layers():
             self.add(layer)
         
     def _check_data(self, x):
@@ -282,11 +318,17 @@ class VQC(tf.keras.Sequential):
             weights = (None, None, None)
         validation_data = (x_val, y_val) if weights is None else (x_val, y_val, weights[1])
 
-        self._attributes['train_size'] = len(x_train)
-        self._attributes['val_size'] = len(x_val)
-        self._attributes['test_size'] = len(x_test)
-        self._attributes['batch_size'] = batch_size
-        self._attributes['epochs'] = epochs   
+        for callback in callbacks:
+            if isinstance(callback, quple.classifiers.VQCLogger):
+                attributes = {}
+                attributes['feature_dimension'] = len(x_train)
+                attributes['train_size'] = len(x_val)
+                attributes['val_size'] =  len(x_val)
+                attributes['test_size'] = len(x_test)
+                attributes['batch_size'] = batch_size
+                attributes['epochs'] = epochs   
+                attributes.update(self.attributes)
+                callback.set_attributes(attributes)
         self.fit(x_train, y_train, batch_size, epochs,
                  validation_data=validation_data,
                  callbacks=callbacks,
