@@ -202,6 +202,19 @@ def parallel_run(func, *iterables, max_workers=None):
 
     return [i for i in result]
 
+def execute_multi_tasks(func, *iterables, parallel):
+    if parallel == 0:
+        result = []
+        for args in zip(*iterables):
+            result.append(func(*args))
+        return result
+    else:
+        if parallel == -1:
+            max_workers = get_cpu_count()
+        else:
+            max_workers = parallel
+        return parallel_run(func, *iterables, max_workers=max_workers)
+
 
 def batching(l:List, n:int):
     for i in range(0, len(l), n):
